#!/usr/bin/env python

"""
A text interface to received and new events.

Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from email import message_from_file
from imiptools import get_handlers, parse_args
from imiptools.config import settings
from imiptools.content import get_objects_from_itip, handle_calendar_data, \
                              handle_calendar_object, have_itip_part, \
                              is_cancel_itip, parse_itip_part
from imiptools.data import get_address, get_periods_using_selector, \
                           get_main_period, get_recurrence_periods, \
                           get_value, parse_object
from imiptools.dates import get_datetime_item, get_time, to_timezone
from imiptools.editing import EditingClient, PeriodError, \
                              form_periods_from_periods
from imiptools.handlers import person, person_outgoing
from imiptools.mail import Messenger
from imiptools.stores import get_journal, get_store
from imiptools.utils import decode_part, message_as_string
import vRecurrence
import sys, os

# User interface functions.

echo = False

def read_input(label):

    """
    Read input, prompting using 'label', stripping leading and trailing
    whitespace, echoing the input if the global 'echo' variable is set.
    """

    s = raw_input(label).strip()
    if echo:
        print s
    return s

def input_with_default(label, default):

    """
    Read input, prompting using 'label', parameterising the label with the given
    'default' and returning the default if no input is given.
    """

    return read_input(label % default) or default

def to_int_or_none(value):

    "Return 'value' as an integer or None for other inputs."

    try:
        return int(value)
    except (TypeError, ValueError):
        return None

def format_value_ranges(ranges, null_value="-"):

    "Format 'ranges' as a single descriptive string."

    l = []

    for value_range in ranges:
        if isinstance(value_range, tuple):
            start, end = value_range
            l.append("%s...%s" % (start, end))
        elif isinstance(value_range, list):
            l.append(", ".join(value_range))
        elif isinstance(value_range, dict):
            l.append(", ".join(value_range.keys()))
        else:
            l.append(value_range or null_value)

    return ", ".join(l)

def print_title(text):

    "Print 'text' with simple, fixed-width styling as a title."

    print text
    print len(text) * "-"

def print_table(rows, separator_index=0):

    """
    Print 'rows' as a simple, fixed-width table. If 'separator_index' is set to
    a row index present in the rows, a table separator will be produced below
    that row's data. Otherwise, a separator will appear below the first row.
    """

    widths = []
    for row in rows:
        for i, col in enumerate(row):
            if i >= len(widths):
                widths.append(len(col))
            else:
                widths[i] = max(widths[i], len(col))

    for i, row in enumerate(rows):
        for col, width in zip(row, widths):
            print "%s%s" % (col, " " * (width - len(col))),
        print
        if i == separator_index:
            for width in widths:
                print "-" * width,
            print

def write(s, filename):

    "Write 's' to a file having the given 'filename'."

    f = filename and open(filename, "w") or None
    try:
        print >>(f or sys.stdout), s
    finally:
        if f:
            f.close()

# Interpret an input file containing a calendar resource.

def get_itip_from_message(filename):

    "Return iTIP details provided by 'filename'."

    f = open(filename)
    try:
        msg = message_from_file(f)
    finally:
        f.close()

    all_itip = []

    for part in msg.walk():
        if have_itip_part(part):
            all_itip.append(parse_itip_part(part))

    return all_itip

def get_itip_from_data(filename, charset):

    "Return objects provided by 'filename'."

    f = open(filename)
    try:
        itip = parse_object(f, charset, "VCALENDAR")
    finally:
        f.close()

    return [itip]

# Object and request display.

def show_objects(objects, user, store):

    """
    Show details of 'objects', accessed by the given 'user' in the given
    'store'.
    """

    print
    print_title("Objects")
    print

    for index, obj in enumerate(objects):
        recurrenceid = obj.get_recurrenceid()
        recurrence_label = recurrenceid and " %s" % recurrenceid or ""
        print "(%d) Summary: %s (%s%s)" % (index, obj.get_value("SUMMARY"), obj.get_uid(), recurrence_label)

def show_requests(user, store):

    "Show requests available to the given 'user' in the given 'store'."

    requests = store.get_requests(user)

    print
    print_title("Requests")
    print

    if not requests:
        print "No requests are pending."
        return

    for index, (uid, recurrenceid) in enumerate(requests):
        obj = store.get_event(user, uid, recurrenceid)
        recurrence_label = recurrenceid and " %s" % recurrenceid or ""
        print "(%d) Summary: %s (%s%s)" % (index, obj.get_value("SUMMARY"), obj.get_uid(), recurrence_label)

# Object details display.

def show_attendee(attendee_item, index):

    "Show the 'attendee_item' (value and attributes) at 'index'."

    attendee, attr = attendee_item
    partstat = attr.get("PARTSTAT")
    print "(%d) %s%s" % (index, attendee, partstat and " (%s)" % partstat or "")

def show_attendees_raw(attendee_map):

    "Show the 'attendee_map' in a simple raw form."

    for attendee, attr in attendee_map.items():
        print attendee

def show_periods(periods, errors=None):

    "Show 'periods' with any indicated 'errors'."

    main = get_main_period(periods)
    if main:
        show_period(main, 0, errors)

    recurrences = get_recurrence_periods(periods)
    if recurrences:
        print
        print_title("Recurrences")
        for index, p in enumerate(recurrences):
            show_period(p, index + 1, errors)

def show_period(p, index, errors=None):

    "Show period 'p' at 'index' with any indicated 'errors'."

    errors = errors and errors.get(index)
    if p.cancelled:
        label = "Cancelled"
    elif p.replacement:
        label = "Replaced"
    elif p.new_replacement:
        label = "To replace"
    elif p.recurrenceid:
        label = "Retained"
    else:
        label = "New"

    error_label = errors and " (errors: %s)" % ", ".join(errors) or ""
    print "(%d) %s%s:" % (index, label, error_label), p.get_start(), p.get_end(), p.origin

def show_periods_raw(periods):

    "Show 'periods' in a simple raw form."

    periods = periods[:]
    periods.sort()
    map(show_period_raw, periods)

def show_period_raw(p):

    "Show period 'p' in a simple raw form."

    print p.get_start(), p.get_end(), p.origin

def show_rule(selectors):

    "Show recurrence rule specification 'selectors'."

    # Collect limit, selection and frequency details.

    for i, selector in enumerate(selectors):
        prefix = "(%d) " % i
        show_rule_selector(selector, prefix)

    print
    print vRecurrence.to_string(selectors)

def show_rule_selector(selector, prefix=""):

    "Show the rule 'selector', employing any given 'prefix' for formatting."

    # COUNT

    if isinstance(selector, vRecurrence.LimitSelector):
        print "%sAt most %d occurrences" % (prefix, selector.args["values"][0])

    # BYSETPOS

    elif isinstance(selector, vRecurrence.PositionSelector):
        for value in selector.get_positions():
            print "%sSelect occurrence #%d" % (prefix, value)
            prefix = len(prefix) * " "

    # BYWEEKDAY

    elif isinstance(selector, vRecurrence.WeekDayFilter):
        for value, index in selector.get_values():
            print "%sSelect occurrence #%d (from %s) of weekday %s" % (
                prefix, abs(index), index >= 0 and "start" or "end",
                get_weekday(value))
            prefix = len(prefix) * " "

    # BY...

    elif isinstance(selector, vRecurrence.Enum):
        for value in selector.get_values():
            print "%sSelect %s %r" % (prefix, get_resolution(selector.level),
                                      value)
            prefix = len(prefix) * " "

    # YEARLY, MONTHLY, WEEKLY, DAILY, HOURLY, MINUTELY, SECONDLY

    elif isinstance(selector, vRecurrence.Pattern):
        print "%sEach %s with interval %d" % (prefix,
            get_resolution(selector.level), selector.args.get("interval", 1))

def get_resolution(level):

    "Return a textual description of the given resolution 'level'."

    levels = ["year", "month", "week", "day in year", "day in month", "day", "hour", "minute", "second"]
    return levels[level]

def get_weekday(weekday):

    "Return the name of the given 1-based 'weekday' number."

    weekdays = [None, "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"]
    return weekdays[weekday]

def show_attendee_changes(new, modified, unmodified, removed):

    "Show 'new', 'modified', 'unmodified' and 'removed' periods."

    print
    print_title("Changes to attendees")
    print
    print "New:"
    show_attendees_raw(new)
    print
    print "Modified:"
    show_attendees_raw(modified)
    print
    print "Unmodified:"
    show_attendees_raw(unmodified)
    print
    print "Removed:"
    show_attendees_raw(removed)

def show_period_classification(new, replaced, retained, cancelled, obsolete):

    "Show 'new', 'replaced', 'retained', 'cancelled' and 'obsolete' periods."

    print
    print_title("Period classification")
    print
    print "New:"
    show_periods_raw(new)
    print
    print "Replaced:"
    show_periods_raw(replaced)
    print
    print "Retained:"
    show_periods_raw(retained)
    print
    print "Cancelled:"
    show_periods_raw(cancelled)
    print
    print "Obsolete:"
    show_periods_raw(obsolete)

def show_changes(modified, unmodified, removed):

    "Show 'modified', 'unmodified' and 'removed' periods."

    print
    print_title("Changes to periods")
    print
    print "Modified:"
    show_periods_raw(modified)
    print
    print "Unmodified:"
    show_periods_raw(unmodified)
    print
    print "Removed:"
    show_periods_raw(removed)

def show_attendee_operations(to_invite, to_cancel, to_modify):

    "Show attendees 'to_invite', 'to_cancel' and 'to_modify'."

    print
    print_title("Attendee update operations")
    print
    print "To invite:"
    show_attendees_raw(to_invite)
    print
    print "To cancel:"
    show_attendees_raw(to_cancel)
    print
    print "To modify:"
    show_attendees_raw(to_modify)

def show_period_operations(to_unschedule, to_reschedule, to_add, to_exclude, to_set,
                           all_unscheduled, all_rescheduled):

    """
    Show operations for periods 'to_unschedule', 'to_reschedule', 'to_add',
    'to_exclude' and 'to_set' (for updating other calendar participants), and
    for periods 'all_unscheduled' and 'all_rescheduled' (for publishing event
    state).
    """

    print
    print_title("Period update and publishing operations")
    print
    print "Unschedule:"
    show_periods_raw(to_unschedule)
    print
    print "Reschedule:"
    show_periods_raw(to_reschedule)
    print
    print "Added:"
    show_periods_raw(to_add)
    print
    print "Excluded:"
    show_periods_raw(to_exclude)
    print
    print "Set in object:"
    show_periods_raw(to_set)
    print
    print "All unscheduled:"
    show_periods_raw(all_unscheduled)
    print
    print "All rescheduled:"
    show_periods_raw(all_rescheduled)

class TextClient(EditingClient):

    "Simple client with textual output."

    def new_object(self):

        "Create a new object with the current time."

        utcnow = get_time()
        now = to_timezone(utcnow, self.get_tzid())
        obj = EditingClient.new_object(self, "VEVENT")
        obj.set_value("SUMMARY", "New event")
        obj["DTSTART"] = [get_datetime_item(now)]
        obj["DTEND"] = [get_datetime_item(now)]
        return obj

    def handle_outgoing_object(self):

        "Handle the current object using the outgoing handlers."

        unscheduled_objects, rescheduled_objects, added_objects = \
            self.get_publish_objects()

        handlers = get_handlers(self, person_outgoing.handlers,
                                [get_address(self.user)])

        # Handle the parent object plus any altered periods.

        handle_calendar_object(self.obj, handlers, "PUBLISH")

        for o in unscheduled_objects:
            handle_calendar_object(o, handlers, "CANCEL")

        for o in rescheduled_objects:
            handle_calendar_object(o, handlers, "PUBLISH")

        for o in added_objects:
            handle_calendar_object(o, handlers, "ADD")

    def update_periods_from_rule(self):

        "Update the periods from the rule."

        selectors = self.state.get("rule")
        periods = self.state.get("periods")

        main_period = get_main_period(periods)
        tzid = main_period.get_tzid()

        start = main_period.get_start()
        end = self.get_window_end() or None

        selector = vRecurrence.get_selector(start, selectors)
        until = None
        inclusive = False

        # Generate the periods from the rule.

        rule_periods = form_periods_from_periods(
                       get_periods_using_selector(selector, main_period, tzid,
                                                  start, end, inclusive))

        # Retain any applicable replacement periods that either modify or cancel
        # rule periods.
        # NOTE: To be done.

        self.state.set("periods", rule_periods)

    # Editing methods involving interaction.

    def add_rule_selectors(self):

        "Add rule selectors to the rule details."

        selectors = self.state.get("rule")

        while True:

            # Obtain a command from any arguments.

            s = read_input("Selector: (c)ount, (f)requency, (s)election (or return)> ")
            args = s.split()
            cmd = next_arg(args)

            if cmd in ("c", "count", "limit"):
                add_rule_selector_count(selectors, args)
            elif cmd in ("f", "freq", "frequency"):
                add_rule_selector_frequency(selectors, args)
            elif cmd in ("s", "select", "selection"):
                add_rule_selector_selection(selectors, args)

            # Remain in the loop unless explicitly terminated.

            elif not cmd or cmd == "end":
                break

    def edit_attendee(self, index):

        "Edit the attendee at 'index'."

        t = self.can_edit_attendee(index)
        if t:
            attendees = self.state.get("attendees")
            attendee, attr = t
            del attendees[attendee]
            attendee = input_with_default("Attendee (%s)? ", attendee)
            attendees[attendee] = attr

    def edit_period(self, index, args=None):

        "Edit the period at 'index'."

        period = self.can_edit_period(index)
        if period:
            edit_period(period, args)
            period.cancelled = False

            # Change the origin of modified rule periods.

            if period.origin == "RRULE":
                period.origin = "RDATE"

            # Sort the periods after this change.

            periods = self.state.get("periods")
            periods.sort()

    def edit_rule_selector(self, index, args):

        "Edit the selector having the given 'index'."

        selectors = self.state.get("rule")
        selector = self.can_edit_rule_selector(index)

        if not selector:
            return

        while True:
            show_rule_selector(selector)

            # Obtain a command from any arguments.

            cmd = next_arg(args)
            if not cmd:
                s = read_input("Selector: (e)dit, (r)emove (or return)> ")
                args = s.split()
                cmd = next_arg(args)

            # Edit an existing selector.

            if cmd in ("e", "edit"):
                if isinstance(selector, vRecurrence.LimitSelector):
                    add_rule_selector_count(selectors, args, selector)
                elif isinstance(selector, vRecurrence.Pattern):
                    add_rule_selector_frequency(selectors, args, selector)
                else:
                    add_rule_selector_selection(selectors, args, selector)

            # Remove an existing selector.

            elif cmd in ("r", "remove"):
                del selectors[index]

            # Exit if requested or after a successful
            # operation.

            elif not cmd:
                pass
            else:
                continue
            break

    def edit_summary(self, summary=None):

        "Edit or set the 'summary'."

        if self.can_edit_properties():
            if not summary:
                summary = input_with_default("Summary (%s)? ", self.state.get("summary"))
            self.state.set("summary", summary)

    def finish(self):

        "Finish editing, warning of errors if any occur."

        try:
            EditingClient.finish(self)
        except PeriodError:
            print "Errors exist in the periods."
            return

    # Diagnostic methods.

    def show_period_classification(self):

        "Show the classification of the periods."

        try:
            new, replaced, retained, cancelled, obsolete = self.classify_periods()
            show_period_classification(new, replaced, retained, cancelled, obsolete)
        except PeriodError:
            print
            print "Errors exist in the periods."

    def show_changes(self):

        "Show how the periods have changed."

        try:
            modified, unmodified, removed = self.classify_period_changes()
            show_changes(modified, unmodified, removed)
        except PeriodError:
            print "Errors exist in the periods."

        is_changed = self.properties_changed()
        if is_changed:
            print
            print "Properties changed:", ", ".join(is_changed)
        new, modified, unmodified, removed = self.classify_attendee_changes()
        show_attendee_changes(new, modified, unmodified, removed)

    def show_operations(self):

        "Show the operations required to change the periods for recipients."

        is_changed = self.properties_changed()

        try:
            to_unschedule, to_reschedule, to_add, to_exclude, to_set, \
                all_unscheduled, all_rescheduled = self.classify_period_operations()
            show_period_operations(to_unschedule, to_reschedule, to_add,
                                   to_exclude, to_set,
                                   all_unscheduled, all_rescheduled)
        except PeriodError:
            print "Errors exist in the periods."

        to_invite, to_cancel, to_modify = self.classify_attendee_operations()
        show_attendee_operations(to_invite, to_cancel, to_modify)

    # Output methods.

    def show_message(self, message, plain=False, filename=None):

        """
        Show the given mail 'message', decoding to plain text if 'plain' is set
        to a true value, writing it to 'filename' if indicated.
        """

        if plain:
            decode_part(message)
        write(message_as_string(message), filename)

    def show_cancel_message(self, plain=False, filename=None):

        "Show the cancel message for uninvited attendees."

        message = self.prepare_cancel_message()
        if message:
            self.show_message(message, plain, filename)

    def show_publish_message(self, plain=False, filename=None):

        "Show the publishing message for the updated event."

        message = self.prepare_publish_message()
        self.show_message(message, plain, filename)

    def show_update_message(self, plain=False, filename=None):

        "Show the update message for the updated event."

        message = self.prepare_update_message()
        if message:
            self.show_message(message, plain, filename)

    # General display methods.

    def show_object(self):
        print
        print_title("Object details")
        print
        print "Summary:", self.state.get("summary")
        print
        print "Organiser:", self.state.get("organiser")
        self.show_attendees()
        self.show_periods()
        self.show_rule()
        self.show_suggested_attendees()
        self.show_suggested_periods()
        self.show_conflicting_periods()
        print
        print "Object is", self.obj.is_shared() and "shared" or "not shared"

    def show_attendees(self):
        print
        print_title("Attendees")
        attendees = self.state.get("attendees")
        for index, attendee_item in enumerate(attendees.items()):
            show_attendee(attendee_item, index)

    def show_periods(self):
        print
        print_title("Periods")
        show_periods(self.state.get("periods"), self.state.get("period_errors"))

    def show_rule(self):
        selectors = self.state.get("rule")
        if selectors:
            print
            print_title("Period recurrence rule")
            show_rule(selectors)

    def show_suggested_attendees(self):
        current_attendee = None
        for index, (attendee, suggested_item) in enumerate(self.state.get("suggested_attendees")):
            if attendee != current_attendee:
                print
                print_title("Attendees suggested by %s" % attendee)
                current_attendee = attendee
            show_attendee(suggested_item, index)

    def show_suggested_periods(self):
        periods = self.state.get("suggested_periods")
        current_attendee = None
        index = 0
        for attendee, period, operation in periods:
            if attendee != current_attendee:
                print
                print_title("Periods suggested by %s" % attendee)
                current_attendee = attendee
            show_period(period, index)
            print "  %s" % (operation == "add" and "Add this period" or "Remove this period")
            index += 1

    def show_conflicting_periods(self):
        conflicts = self.get_conflicting_periods()
        if not conflicts:
            return
        print
        print_title("Conflicting periods")

        conflicts = list(conflicts)
        conflicts.sort()

        for p in conflicts:
            print p.summary, p.uid, p.get_start(), p.get_end()

# Interaction functions.

def expand_arg(args):

    """
    Expand the first argument in 'args' to a pair of arguments if having the
    form <char><digit>...
    """

    if args[0] and args[0][1:].isdigit():
        args[:1] = [args[0][0], args[0][1:]]

def get_text_arg(s):

    """
    Split 's' after the first whitespace occurrence, returning the remaining
    text or None if no such text exists.
    """

    return (s.split(None, 1)[1:] or [None])[0]

def next_arg(args):

    """
    Return the first argument from 'args', removing it, or return None if no
    arguments are left.
    """

    if args:
        arg = args[0]
        del args[0]
        return arg
    return None

# Editing functions.

def edit_period(period, args=None):

    "Edit the given 'period'."

    print "Editing start (%s)" % period.get_start()
    edit_date(period.start, args)
    print "Editing end (%s)" % period.get_end()
    edit_date(period.end, args)

def edit_date(date, args=None):

    "Edit the given 'date' object attributes."

    date.date = next_arg(args) or input_with_default("Date (%s)? ", date.date)
    date.hour = next_arg(args) or input_with_default("Hour (%s)? ", date.hour)
    date.minute = next_arg(args) or input_with_default("Minute (%s)? ", date.minute)
    date.second = next_arg(args) or input_with_default("Second (%s)? ", date.second)
    date.tzid = next_arg(args) or input_with_default("Time zone (%s)? ", date.tzid)
    date.reset()

def add_rule_selector_count(selectors, args, selector=None):

    "Add to 'selectors' a selector imposing a count restriction."

    while True:
        arg = next_arg(args)
        if not arg:
            if selector:
                arg = input_with_default("Number of occurrences (%d)? ",
                                         selector.get_limit())
            else:
                arg = read_input("Number of occurrences? ")

        count = to_int_or_none(arg)

        if count is None:
            arg = None
            continue

        # Change or add selector.

        selector = selector or selectors and \
                   isinstance(selectors[0], vRecurrence.LimitSelector) and \
                   selectors[0] or None

        if not selector:
            selector = vRecurrence.new_selector("COUNT")
            selectors.insert(0, selector)

        selector.set_limit(count)
        break

def add_rule_selector_frequency(selectors, args, selector=None):

    "Add to 'selectors' a selector for a frequency."

    while not selector:
        arg = next_arg(args)
        if not arg:
            arg = read_input("Select (y)early, (M)onthly, (w)eekly, (d)aily, "
                             "(h)ourly, (m)inutely, (s)econdly (or return)? ")

        if not arg:
            return

        arg_lower = arg.lower()

        if arg_lower in ("y", "year", "yearly"):
            qualifier = "YEARLY"
        elif arg == "M" or arg_lower in ("month", "monthly"):
            qualifier = "MONTHLY"
        elif arg_lower in ("w", "week", "weekly"):
            qualifier = "WEEKLY"
        elif arg_lower in ("d", "day", "daily"):
            qualifier = "DAILY"
        elif arg_lower in ("h", "hour", "hourly"):
            qualifier = "HOURLY"
        elif arg_lower in ("m", "minute", "minutely"):
            qualifier = "MINUTELY"
        elif arg_lower in ("s", "second", "secondly"):
            qualifier = "SECONDLY"
        else:
            continue
 
        break

    while True:
        arg = next_arg(args)
        if not arg:
            if selector:
                arg = input_with_default("Interval (%d)? ",
                                         selector.get_interval())
            else:
                arg = input_with_default("Interval (%d)? ", 1)

        interval = to_int_or_none(arg)

        if interval is None:
            arg = None
        else:
            break

    # Update an existing selector.

    if selector:
        selector.set_interval(interval)
        return

    # Create a new selector.

    selector = vRecurrence.new_selector(qualifier)
    selector.set_interval(interval)

    # Remove any existing frequency selector.

    for index, _selector in enumerate(selectors):
        if isinstance(_selector, vRecurrence.Pattern):
            del selectors[index]
            break

    # Add the new selector and keep the selectors in order.

    selectors.append(selector)
    vRecurrence.sort_selectors(selectors)

def add_rule_selector_selection(selectors, args, selector=None):

    "Add to 'selectors' a selector for a particular point in time."

    qualifier = selector and selector.qualifier or None

    while not selector:
        arg = next_arg(args)
        if not arg:
            arg = read_input("Select (M)onths, (w)eeks, (y)eardays, "
                             "m(o)nthdays, week(d)ays, (h)ours, (m)inutes, "
                             "(s)econds (or return)? ")

        if not arg:
            return

        arg_lower = arg.lower()

        if arg == "M" or arg_lower in ("month", "months"):
            qualifier = "BYMONTH"
        elif arg_lower in ("w", "week", "weeks"):
            qualifier = "BYWEEKNO"
        elif arg_lower in ("y", "yearday", "yeardays"):
            qualifier = "BYYEARDAY"
        elif arg_lower in ("o", "monthday", "monthdays"):
            qualifier = "BYMONTHDAY"
        elif arg_lower in ("d", "weekday", "weekdays"):
            qualifier = "BYDAY"
        elif arg_lower in ("h", "hour", "hours"):
            qualifier = "BYHOUR"
        elif arg_lower in ("m", "minute", "minutes"):
            qualifier = "BYMINUTE"
        elif arg_lower in ("s", "second", "seconds"):
            qualifier = "BYSECOND"
        else:
            continue
 
        break

    if not qualifier:
        return

    ranges = vRecurrence.get_value_ranges(qualifier)
    ranges_str = format_value_ranges(ranges[0])

    values = []

    while True:
        arg = next_arg(args)
        if not arg:
            arg = read_input("Value (%s) (return to end)? " % ranges_str)

        # Stop if no more arguments.

        if not arg or arg == "end":
            break

        # Handle weekdays.

        if qualifier == "BYDAY":
            value = arg.upper() # help to match weekdays

            arg = next_arg(args)
            if not arg:
                arg = read_input("Occurrence within a month? ")

            index = to_int_or_none(arg)
            value = vRecurrence.check_values(qualifier, [value, index])

        # Handle all other values.

        else:
            value = to_int_or_none(arg)
            l = vRecurrence.check_values(qualifier, [value])
            value = l and l[0]

        # Append valid values.

        if value is not None:
            values.append(value)
        else:
            print "Value not recognised."

    if not values:
        return

    # Update an existing selector.

    if selector:
        selector.set_values(values)
        return

    # Create a new selector.

    selector = vRecurrence.new_selector(qualifier)
    selector.set_values(values)

    # Remove any existing selector.

    for index, _selector in enumerate(selectors):
        if _selector.qualifier == selector.qualifier:
            del selectors[index]
            break

    # Add the new selector and keep the selectors in order.

    selectors.append(selector)
    vRecurrence.sort_selectors(selectors)

def select_object(cl, objects):

    "Select using 'cl' an object from the given 'objects'."

    print

    if objects:
        label = "Select object number or (n)ew object or (q)uit> "
    else:
        label = "Select (n)ew object or (q)uit> "

    while True:
        try:
            cmd = read_input(label)
        except EOFError:
            return None

        if cmd.isdigit():
            index = to_int_or_none(cmd)

            if index is not None and 0 <= index < len(objects):
                obj = objects[index]
                return cl.load_object(obj.get_uid(), obj.get_recurrenceid())

        elif cmd in ("n", "new"):
            return cl.new_object()
        elif cmd in ("q", "quit", "exit"):
            return None

def show_commands():

    "Show editing and inspection commands."

    print
    print_title("Editing commands")
    print
    print """\
a [ <uri> ]
attendee [ <uri> ]
   Add attendee

A, attend, attendance
   Change attendance/participation

a<digit>
attendee <digit>
   Select attendee from list

as<digit>
   Add suggested attendee from list

f, finish
   Finish editing, confirming changes, proceeding to messaging

h, help, ?
   Show this help message

l, list, show
   List/show all event details

p, period [ new ]
   Add new period

p<digit>
period <digit>
   Select period from list

ps<digit>
   Add or remove suggested period from list

q, quit, exit
   Exit/quit this program

r, reload, reset, restart
   Reset event periods (return to editing mode, if already finished)

rule, rrule
   Add a period recurrence rule

rule <digit>
rrule <digit>
   Select period recurrence rule selector from list

s, summary
   Set event summary
"""

    print_title("Messaging commands")
    print
    print """\
S, send
   Send messages to recipients and to self, if appropriate
"""

    print_title("Diagnostic commands")
    print
    print """\
c, class, classification
   Show period classification

C, changes
   Show changes made by editing

o, ops, operations
   Show update operations

RECURRENCE-ID [ <filename> ]
   Show event recurrence identifier, writing to <filename> if specified

UID [ <filename> ]
   Show event unique identifier, writing to <filename> if specified
"""

    print_title("Message inspection commands")
    print
    print """\
P [ <filename> ]
publish [ <filename> ]
   Show publishing message, writing to <filename> if specified

R [ <filename> ]
remove [ <filename> ]
cancel [ <filename> ]
   Show cancellation message sent to uninvited/removed recipients, writing to
   <filename> if specified

U [ <filename> ]
update [ <filename> ]
   Show update message, writing to <filename> if specified
"""

def edit_object(cl, obj, handle_outgoing=False):

    """
    Edit using 'cl' the given object 'obj'. If 'handle_outgoing' is specified
    and set to a true value, the details from outgoing messages are incorporated
    into the stored data.
    """

    cl.show_object()
    print

    try:
        while True:
            role = cl.is_organiser() and "Organiser" or "Attendee"
            status = cl.state.get("finished") and " (editing complete)" or ""

            s = read_input("%s%s> " % (role, status))
            args = s.split()

            if not args or not args[0]:
                continue

            # Expand short-form arguments.

            expand_arg(args)
            cmd = next_arg(args)

            # Check the status of the periods.

            if cmd in ("c", "class", "classification"):
                cl.show_period_classification()
                print

            elif cmd in ("C", "changes"):
                cl.show_changes()
                print

            # Finish editing.

            elif cmd in ("f", "finish"):
                cl.finish()

            # Help.

            elif cmd in ("h", "?", "help"):
                show_commands()

            # Show object details.

            elif cmd in ("l", "list", "show"):
                cl.show_object()
                print

            # Show the operations.

            elif cmd in ("o", "ops", "operations"):
                cl.show_operations()
                print

            # Quit or exit.

            elif cmd in ("q", "quit", "exit"):
                break

            # Restart editing.

            elif cmd in ("r", "reload", "reset", "restart"):
                obj = cl.load_object(obj.get_uid(), obj.get_recurrenceid())
                if not obj:
                    obj = cl.new_object()
                cl.reset()
                cl.show_object()
                print

            # Show UID details.

            elif cmd == "UID":
                filename = get_text_arg(s)
                write(obj.get_uid(), filename)

            elif cmd == "RECURRENCE-ID":
                filename = get_text_arg(s)
                write(obj.get_recurrenceid() or "", filename)

            # Post-editing operations.

            elif cl.state.get("finished"):

                # Show messages.

                if cmd in ("P", "publish"):
                    filename = get_text_arg(s)
                    cl.show_publish_message(plain=not filename, filename=filename)

                elif cmd in ("R", "remove", "cancel"):
                    filename = get_text_arg(s)
                    cl.show_cancel_message(plain=not filename, filename=filename)

                elif cmd in ("U", "update"):
                    filename = get_text_arg(s)
                    cl.show_update_message(plain=not filename, filename=filename)

                # Definitive finishing action.

                elif cmd in ("S", "send"):

                    # Send update and cancellation messages.

                    did_send = False

                    message = cl.prepare_update_message()
                    if message:
                        cl.send_message(message, cl.get_recipients())
                        did_send = True

                    to_cancel = cl.state.get("attendees_to_cancel")
                    if to_cancel:
                        message = cl.prepare_cancel_message()
                        if message:
                            cl.send_message(message, to_cancel)
                            did_send = True

                    # Process the object using the person outgoing handler.

                    if handle_outgoing:
                        cl.handle_outgoing_object()

                    # Otherwise, send a message to self with the event details.

                    else:
                        message = cl.prepare_publish_message()
                        if message:
                            cl.send_message_to_self(message)
                            did_send = True

                    # Exit if sending occurred.

                    if did_send:
                        break
                    else:
                        print "No messages sent. Try making edits or exit manually."

            # Editing operations.

            elif not cl.state.get("finished"):

                # Add or edit attendee.

                if cmd in ("a", "attendee"):
                    value = next_arg(args)
                    index = to_int_or_none(value)

                    if index is None:
                        try:
                            index = cl.find_attendee(value)
                        except ValueError:
                            index = None

                    # Add an attendee.

                    if index is None:
                        cl.add_attendee(value)
                        if not value:
                            cl.edit_attendee(-1)

                    # Edit attendee (using index).

                    else:
                        attendee_item = cl.can_remove_attendee(index)
                        if attendee_item:
                            while True:
                                show_attendee(attendee_item, index)

                                # Obtain a command from any arguments.

                                cmd = next_arg(args)
                                if not cmd:
                                    cmd = read_input("Attendee: (e)dit, (r)emove (or return)> ")
                                if cmd in ("e", "edit"):
                                    cl.edit_attendee(index)
                                elif cmd in ("r", "remove"):
                                    cl.remove_attendees([index])

                                # Exit if requested or after a successful
                                # operation.

                                elif not cmd:
                                    pass
                                else:
                                    continue
                                break

                    cl.show_attendees()
                    print

                # Add suggested attendee (using index).

                elif cmd in ("as", "attendee-suggested", "suggested-attendee"):
                    value = next_arg(args)
                    index = to_int_or_none(value)

                    if index is not None:
                        cl.add_suggested_attendee(index)

                    cl.show_attendees()
                    print

                # Edit attendance.

                elif cmd in ("A", "attend", "attendance"):

                    if not cl.is_attendee() and cl.is_organiser():
                        cl.add_attendee(cl.user)

                    # NOTE: Support delegation.

                    if cl.can_edit_attendance():
                        while True:

                            # Obtain a command from any arguments.

                            cmd = next_arg(args)
                            if not cmd:
                                cmd = read_input("Attendance: (a)ccept, (d)ecline, (t)entative (or return)> ")
                            if cmd in ("a", "accept", "accepted", "attend"):
                                cl.edit_attendance("ACCEPTED")
                            elif cmd in ("d", "decline", "declined"):
                                cl.edit_attendance("DECLINED")
                            elif cmd in ("t", "tentative"):
                                cl.edit_attendance("TENTATIVE")

                            # Exit if requested or after a successful operation.

                            elif not cmd:
                                pass
                            else:
                                continue
                            break

                    cl.show_attendees()
                    print

                # Add or edit period.

                elif cmd in ("p", "period"):
                    value = next_arg(args)
                    index = to_int_or_none(value)

                    # Add a new period.

                    if index is None or value == "new":
                        cl.add_period()
                        cl.edit_period(-1, args)

                    # Edit period (using index).

                    else:
                        period = cl.can_edit_period(index)
                        if period:
                            while True:
                                show_period_raw(period)

                                # Obtain a command from any arguments.

                                cmd = next_arg(args)
                                if not cmd:
                                    cmd = read_input("Period: (e)dit, (c)ancel, (u)ncancel (or return)> ")
                                if cmd in ("e", "edit"):
                                    cl.edit_period(index, args)
                                elif cmd in ("c", "cancel"):
                                    cl.cancel_periods([index])
                                elif cmd in ("u", "uncancel", "restore"):
                                    cl.cancel_periods([index], False)

                                # Exit if requested or after a successful
                                # operation.

                                elif not cmd:
                                    pass
                                else:
                                    continue
                                break

                    cl.show_periods()
                    print

                # Apply suggested period (using index).

                elif cmd in ("ps", "period-suggested", "suggested-period"):
                    value = next_arg(args)
                    index = to_int_or_none(value)

                    if index is not None:
                        cl.apply_suggested_period(index)

                    cl.show_periods()
                    print

                # Specify a recurrence rule.

                elif cmd in ("rule", "rrule"):
                    value = next_arg(args)
                    index = to_int_or_none(value)

                    # Add a new rule.

                    if index is None:
                        cl.add_rule_selectors()
                    else:
                        cl.edit_rule_selector(index, args)

                    cl.show_rule()
                    cl.update_periods_from_rule()
                    print

                # Set the summary.

                elif cmd in ("s", "summary"):
                    cl.edit_summary(get_text_arg(s))
                    cl.show_object()
                    print

    except EOFError:
        return

def main(args):

    """
    The main program, employing command line 'args' to initialise the editing
    activity.
    """

    global echo

    if "--help" in args:
        show_help(os.path.split(sys.argv[0])[-1])
        return 0

    # Parse command line arguments using the standard options plus some extra
    # options.

    args = parse_args(args, {
        "--calendar-data"   : ("calendar_data", False),
        "--charset"         : ("charset", "utf-8"),
        "--echo"            : ("echo", False),
        "-f"                : ("filename", None),
        "--handle-data"     : ("handle_data", False),
        "--suppress-bcc"    : ("suppress_bcc", False),
        "-u"                : ("user", None),
        "--uid"             : ("uid", None),
        "--recurrence-id"   : ("recurrenceid", None),
        "--show-config"     : ("show_config", False)
        })

    charset = args["charset"]
    calendar_data = args["calendar_data"]
    echo = args["echo"]
    filename = args["filename"]
    handle_data = args["handle_data"]
    sender = (args["senders"] or [None])[0]
    suppress_bcc = args["suppress_bcc"]
    user = args["user"]
    uid = args["uid"]
    recurrenceid = args["recurrenceid"]

    # Open a store.

    store_type = args.get("store_type")
    store_dir = args.get("store_dir")
    preferences_dir = args.get("preferences_dir")

    # Show configuration and exit if requested.

    if args["show_config"]:
        print """\
Store type: %s (%s)
Store directory: %s (%s)
Preferences directory: %s
""" % (
            store_type, settings["STORE_TYPE"],
            store_dir, settings["STORE_DIR"],
            preferences_dir)
        return 0

    # Determine the user and sender identities.

    if sender and not user:
        user = get_uri(sender)
    elif user and not sender:
        sender = get_address(user)
    elif not sender and not user:
        print >>sys.stderr, "A sender or a user must be specified."
        return 1

    # Obtain a store but not a journal.

    store = get_store(store_type, store_dir)
    journal = None

    # Open a messenger for the user.

    messenger = Messenger(sender=sender, suppress_bcc=suppress_bcc)

    # Open a client for the user.

    cl = TextClient(user, messenger, store, journal, preferences_dir)

    # Read any input resource, using it to obtain identifier details.

    if filename:
        if calendar_data:
            all_itip = get_itip_from_data(filename, charset)
        else:
            all_itip = get_itip_from_message(filename)

        objects = []

        # Process the objects using the person handler.

        if handle_data:
            for itip in all_itip:
                handled = handle_calendar_data(itip, get_handlers(cl, person.handlers, None))
                if not is_cancel_itip(itip):
                    objects += handled

        # Or just obtain objects from the data.

        else:
            for itip in all_itip:
                handled = get_objects_from_itip(itip, ["VEVENT"])
                if not is_cancel_itip(itip):
                    objects += handled

        # Choose an object to edit.

        show_objects(objects, user, store)
        obj = select_object(cl, objects)

    # Load any indicated object.

    elif uid:
        obj = cl.load_object(uid, recurrenceid)

    # Or create a new object.

    else:
        obj = cl.new_object()

    # Exit without any object.

    if not obj:
        print >>sys.stderr, "No object loaded."
        return 1

    # Edit the object.

    edit_object(cl, obj, handle_outgoing=handle_data)

def show_help(progname):
    print >>sys.stderr, help_text % progname

help_text = """\
Usage: %s -s <sender> | -u <user> \\
         [ -f <filename> | --uid <uid> [ --recurrence-id <recurrence-id> ] ] \\
         [ --calendar-data --charset ] \\
         [ --handle-data ] \\
         [ -T <store type ] [ -S <store directory> ] \\
         [ -p <preferences directory> ] \\
         [ --echo ]

Identity options:

-s  Indicate the user by specifying a sender address
-u  Indicate the user by specifying their URI

Input options:

-f               Indicates a filename containing a MIME-encoded message or
                 calendar object
--uid            Indicates the UID of a stored calendar object
--recurrence-id  Indicates a stored object with a specific RECURRENCE-ID

--calendar-data  Indicates that the specified file contains a calendar object
                 as opposed to a mail message
--charset        Specifies the character encoding used by a calendar object
                 description

Processing options:

--handle-data    Cause the input to be handled and stored in the configured
                 data store

Configuration options (overriding configured defaults):

-p  Indicates the location of user preference directories
-S  Indicates the location of the calendar data store containing user storage
    directories
-T  Indicates the store type (the configured value if omitted)

Output options:

--echo           Echo received input, useful if consuming input from the
                 standard input stream and producing a log of the program's
                 activity
"""

if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))

# vim: tabstop=4 expandtab shiftwidth=4
